"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var sqrt64 = Math.sqrt;
var PI = 3.141592653589793;
var SOLAR_MASS = 4 * PI * PI;
var DAYS_PER_YEAR = 365.24;
var Body = (function () {
    function Body(x, y, z, vx, vy, vz, mass) {
        this.x = x;
        this.y = y;
        this.z = z;
        this.vx = vx;
        this.vy = vy;
        this.vz = vz;
        this.mass = mass;
    }
    Body.prototype.offsetMomentum = function (px, py, pz) {
        this.vx = -px / SOLAR_MASS;
        this.vy = -py / SOLAR_MASS;
        this.vz = -pz / SOLAR_MASS;
        return this;
    };
    return Body;
}());
function Jupiter() {
    return new Body(4.84143144246472090e+00, -1.16032004402742839e+00, -1.03622044471123109e-01, 1.66007664274403694e-03 * DAYS_PER_YEAR, 7.69901118419740425e-03 * DAYS_PER_YEAR, -6.90460016972063023e-05 * DAYS_PER_YEAR, 9.54791938424326609e-04 * SOLAR_MASS);
}
function Saturn() {
    return new Body(8.34336671824457987e+00, 4.12479856412430479e+00, -4.03523417114321381e-01, -2.76742510726862411e-03 * DAYS_PER_YEAR, 4.99852801234917238e-03 * DAYS_PER_YEAR, 2.30417297573763929e-05 * DAYS_PER_YEAR, 2.85885980666130812e-04 * SOLAR_MASS);
}
function Uranus() {
    return new Body(1.28943695621391310e+01, -1.51111514016986312e+01, -2.23307578892655734e-01, 2.96460137564761618e-03 * DAYS_PER_YEAR, 2.37847173959480950e-03 * DAYS_PER_YEAR, -2.96589568540237556e-05 * DAYS_PER_YEAR, 4.36624404335156298e-05 * SOLAR_MASS);
}
function Neptune() {
    return new Body(1.53796971148509165e+01, -2.59193146099879641e+01, 1.79258772950371181e-01, 2.68067772490389322e-03 * DAYS_PER_YEAR, 1.62824170038242295e-03 * DAYS_PER_YEAR, -9.51592254519715870e-05 * DAYS_PER_YEAR, 5.15138902046611451e-05 * SOLAR_MASS);
}
function Sun() {
    return new Body(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, SOLAR_MASS);
}
var NBodySystem = (function () {
    function NBodySystem(bodies) {
        var px = 0.0;
        var py = 0.0;
        var pz = 0.0;
        var size = bodies.length;
        for (var i = 0; i < size; i++) {
            var b = bodies[i];
            var m = b.mass;
            px += b.vx * m;
            py += b.vy * m;
            pz += b.vz * m;
        }
        this.bodies = bodies;
        this.bodies[0].offsetMomentum(px, py, pz);
    }
    NBodySystem.prototype.advance = function (dt) {
        var dx, dy, dz, ix, iy, iz, bivx, bivy, bivz, distance, mag;
        var bodies = this.bodies;
        var size = bodies.length;
        for (var i = 0; i < size; ++i) {
            var bodyi = bodies[i];
            ix = bodyi.x;
            iy = bodyi.y;
            iz = bodyi.z;
            bivx = bodyi.vx;
            bivy = bodyi.vy;
            bivz = bodyi.vz;
            var bodyim = bodyi.mass;
            for (var j = i + 1; j < size; ++j) {
                var bodyj = bodies[j];
                dx = ix - bodyj.x;
                dy = iy - bodyj.y;
                dz = iz - bodyj.z;
                var distanceSq = dx * dx + dy * dy + dz * dz;
                distance = sqrt64(distanceSq);
                mag = dt / (distanceSq * distance);
                var bim = bodyim * mag;
                var bjm = bodyj.mass * mag;
                bivx -= dx * bjm;
                bivy -= dy * bjm;
                bivz -= dz * bjm;
                bodyj.vx += dx * bim;
                bodyj.vy += dy * bim;
                bodyj.vz += dz * bim;
            }
            bodyi.vx = bivx;
            bodyi.vy = bivy;
            bodyi.vz = bivz;
            bodyi.x += dt * bivx;
            bodyi.y += dt * bivy;
            bodyi.z += dt * bivz;
        }
    };
    NBodySystem.prototype.energy = function () {
        var dx, dy, dz, distance;
        var ix, iy, iz, vx, vy, vz, bim;
        var e = 0.0;
        var bodies = this.bodies;
        var size = bodies.length;
        for (var i = 0; i < size; ++i) {
            var bodyi = bodies[i];
            ix = bodyi.x;
            iy = bodyi.y;
            iz = bodyi.z;
            vx = bodyi.vx;
            vy = bodyi.vy;
            vz = bodyi.vz;
            bim = bodyi.mass;
            e += 0.5 * bim * (vx * vx + vy * vy + vz * vz);
            for (var j = i + 1; j < size; ++j) {
                var bodyj = bodies[j];
                dx = ix - bodyj.x;
                dy = iy - bodyj.y;
                dz = iz - bodyj.z;
                distance = sqrt64(dx * dx + dy * dy + dz * dz);
                e -= bim * bodyj.mass / distance;
            }
        }
        return e;
    };
    return NBodySystem;
}());
function test(n) {
    var bodies = new NBodySystem(new Array(Sun(), Jupiter(), Saturn(), Uranus(), Neptune()));
    for (var i = 0; i < n; i++) {
        bodies.advance(0.01);
    }
    return bodies.energy();
}
exports.test = test;
